/* 
 * Fast one-pass Floyd Steinberg error diffusion for 8 bit GIF type palette images.
 * Paul Ausbeck 7/3/98
 * Copyright 1998 by Paul Ausbeck
 */

#define _INC_DITHER

#ifndef _INC_IFILELIB
#include "ifilelib.h"
#endif

#define INT short

class color
{
  INT r;
  INT g;
  INT b;

  friend class color_quantizer8;
  friend class color_cube8x6;

public:
  color(void) { r = g = b = 0; }
  color(color &src) { r = src.r; g = src.g; b = src.b;}
  color(int index, ifile_palette_entry *pal)
  {
    r = (pal + index)->red;
    g = (pal + index)->green;
    b = (pal + index)->blue;
  }
  color(ifile_palette_entry *pal)
  {
    r = pal->red; g = pal->green; b = pal->blue;
  }

  color &operator =(int set){ r = g = b = set; return *this;}

  color &operator +=(color &c) { r += c.r; g += c.g; b += c.b; return *this; }
  color &operator -=(color &c) { r -= c.r; g -= c.g; b -= c.b; return *this; }
  color &operator <<=(int shift) 
  { 
    r <<= shift; g <<= shift; b <<= shift;
    return *this;
  }
  color &operator >>=(int shift) 
  { 
    r >>= shift; g >>= shift; b >>= shift;
    return *this;
  }

  void clamp(INT max)
  {
    r = r < 0 ? 0 : (r > max ? max : r);
    g = g < 0 ? 0 : (g > max ? max : g);
    b = b < 0 ? 0 : (b > max ? max : b);
  }
  void clamp(void)
  {
    r = r < 0 ? 0 : (r > 255 ? 255 : r);
    g = g < 0 ? 0 : (g > 255 ? 255 : g);
    b = b < 0 ? 0 : (b > 255 ? 255 : b);
  }
};

class color_quantizer8 // 8 bit color depth.
{
  unsigned char q[256];

public:

  color_quantizer8(int levels = 6)
  {
    int size = levels*levels*levels;
    int bucket = 256/(levels - 1);
    int half = bucket/2;

    for (int index = 0; index < 256; index++)
    {
      q[index] = (index + half)/bucket;
    }
  }
  ~color_quantizer8(void) { }

  int quantize(int value) { return *(q + value); }
  color &quantize(color &value)
  {
    value.r = *(q + value.r);
    value.g = *(q + value.g);
    value.b = *(q + value.b);

    return value;
  }
};

class color_cube8x6 // depth 8 and 6 quantization levels
{
  ifile_palette_entry *pal;
  static INT six[6];
  static INT thirty_six[6];

  enum numbers {palette_size = 216};

public:

  color_cube8x6(void)
  {
    pal = new ifile_palette_entry[palette_size];

    for (int b = 0; b < 6; b++)
      for (int g = 0; g < 6; g++)
        for (int r = 0; r < 6; r++)
        {
          pal[index(r, g, b)].red = r*51; 
          pal[index(r, g, b)].green = g*51; 
          pal[index(r, g, b)].blue = b*51; 
        }
  }
  ~color_cube8x6(void) { delete[] pal; }

  ifile_palette_entry *palette(void) { return pal; }
  int colors(void) { return palette_size; }

  INT index(INT r, INT g, INT b)
  {
    return r + six[g] + thirty_six[b];
  }
  INT index(color &c)
  {
    return c.r + six[c.g] + thirty_six[c.b];
  }
  void expand(INT index, color &c)
  {
    ifile_palette_entry *temp = pal + index;
    c.r = temp->red; c.g = temp->green; c.b = temp->blue;
  }
  INT expand(color &c)
  {
    INT index = c.r + six[c.g] + thirty_six[c.b];
    ifile_palette_entry *temp = pal + index;
    c.r = temp->red; c.g = temp->green; c.b = temp->blue;
    return index;
  }
};

class fsdither8x6 // depth 8 and 6 quantization levels.
{
  static color_quantizer8 *sq;
  static color_cube8x6 *scube;
  
  color_quantizer8 *q;
  color_cube8x6 *cube;
  ifile_palette_entry *pal;

  int width;
  color *errors;

  int transparent;

public:

  fsdither8x6(ifile_palette_entry *pal, int width)
  { 
    this->width = width;
    this->pal = pal;
     
    cube = scube;
    q = sq;

    errors = new color[(width + 2)];  
  }
  ~fsdither8x6(void) { delete[] errors;}

  void dither_row(unsigned char *in, unsigned char *out);

  void dither(unsigned char* in, unsigned char *out, int stride, int rows, 
              int transparent = -1);

  ifile_palette_entry *palette(void) { return cube->palette(); }
  int colors(void) { return cube->colors(); }

  static void init(void);
};